import http from "http";
import https from "https";
import { URL } from "url";
import querystring from "querystring";
import zlib from "zlib";

type ResponseData = {
  body:        http.IncomingMessage;
  headers:     http.IncomingHttpHeaders;
  text:        () => Promise<string>;
  json:        () => Promise<JSON>;
  buffer:      () => Promise<Buffer>;
  arrayBuffer: () => Promise<ArrayBuffer>;
};

interface ExtendedRequestOptions extends http.RequestOptions {
  body?: any;
  timeout?: number;
}

const decompress = (
  data: Buffer,
  encoding: string | undefined
): Buffer => {
  switch(encoding) {
    case "br":      return zlib.brotliDecompressSync(data);
    case "gzip":    return zlib.gunzipSync(data);
    case "deflate": return zlib.inflateSync(data);
  }
  return data;
};

const readData = <T>(
  res: http.IncomingMessage,
  mode: "text" | "json" | "buffer" | "arraybuffer"
): Promise<T> =>
  new Promise((resolve, reject) => {
    const chunks: Buffer[] = [];
    res
      .on("data", chunk => chunks.push(chunk))
      .on("end", () => {
        try {
          const data = decompress(Buffer.concat(chunks), res.headers["content-encoding"]);
          if(mode === "json")
            resolve(JSON.parse(data.toString("utf8")));
          else if(mode === "buffer")
            resolve(data as T);
          else if(mode === "arraybuffer")
            resolve(new Uint8Array(data).buffer as T);
          else
            resolve(data.toString("utf8") as T);
        }
        catch(err: any) {
          reject(err);
        }
      })
      .on("error", reject);
  });

export default async (
  urlString: string,
  options: ExtendedRequestOptions = {}
): Promise<ResponseData> => {
  const { protocol, hostname, pathname, search, port } = new URL(urlString);

  const body =
    options.method === "POST" && options.body
      ? options.headers?.["Content-Type"] === "application/json"
        ? JSON.stringify(options.body)
        : querystring.stringify(options.body as querystring.ParsedUrlQueryInput)
      : null;

  const requestOptions: http.RequestOptions = {
    hostname,
    port: port || (protocol === "https:" ? 443 : 80),
    path: pathname + search,
    method: options.method || "GET",
    headers: {
      ...options.headers,
      ...(body ? { "Content-Length": Buffer.byteLength(body) } : {}),
      "Accept-Encoding": "br, gzip, deflate"
    }
  };

  return new Promise((resolve, reject) => {
    const requester = protocol === "https:" ? https : http;
    const req = requester.request(requestOptions, res => {
      if(res.statusCode && (res.statusCode < 200 || res.statusCode >= 300))
        return reject(new Error(`Request failed with status code ${res.statusCode}`));
      resolve({
        body:        res,
        headers:     res.headers,
        text:        () => readData<string>(res, "text"),
        json:        () => readData<JSON>(res, "json"),
        buffer:      () => readData<Buffer>(res, "buffer"),
        arrayBuffer: () => readData<ArrayBuffer>(res, "arraybuffer")
      });
    });

    req.setTimeout(options.timeout || 5e3, () => {
      req.destroy();
      reject(new Error("Request timed out"));
    });

    req.on("error", reject);
    if(body)
      req.write(body);
    req.end();
  });
};
